#include <stdlib.h>
#include <stdio.h>
#include <math.h>

#include "svf.h"


#define CIRCLE_Z 0.2
#define COONS_STEPS 10
#define PI 3.141592654

#define max(A,B) (((A)>(B)) ? A : B)
#define min(A,B) (((A)<(B)) ? A : B)

int Fround(float f)
{
  if ((f-(int)f)<0.5) {return floor(f);} else return ceil(f);
}


void draw_absolute_line(float x1,float y1,float x2,float y2)

{
  int k,k1,k2,dx,dy,x,y,d;

  dx=abs(Fround(x2)-Fround(x1));
  dy=abs(Fround(y2)-Fround(y1));
  if (dx>=dy)
  {
    if (x1>x2) {x=Fround(x1);y=Fround(y1);x1=x2;y1=y2;x2=x;y2=y;}
    k1=2*dy;
    k2=2*(dy-dx);
    d=2*Fround(dy)-Fround(dx);
    if (y1>y2) k=-1; else k=1;
    x=Fround(x1); y=Fround(y1);
    putpixel(x,y);
    if (Actual.width>1) putpixel(x,y+1);
    if (Actual.width>2) putpixel(x,y-1);
    while (x<Fround(x2))
    {
      x++;
      if (d>=0) {y+=k; d=d+k2;}
      else d=d+k1;
      putpixel(x,y);
      if (Actual.width>1) putpixel(x,y+1);
      if (Actual.width>2) putpixel(x,y-1);
    }
  }
  else
  {
    if (y1>y2) {x=Fround(x1);y=Fround(y1);x1=x2;y1=y2;x2=x;y2=y;}
    k1=2*dx;
    k2=2*(dx-dy);
    d=2*dx-dy;
    if (x2<x1) k=-1; else k=1;
    x=Fround(x1); y=Fround(y1);
    putpixel(x,y);
    if (Actual.width>1) putpixel(x+1,y);
    if (Actual.width>2) putpixel(x-1,y);
    while (y<Fround(y2))
    {
      y++;
      if (d>=0) {x+=k; d=d+k2;}
      else d=d+k1;
      putpixel(x,y);
      if (Actual.width>1) putpixel(x+1,y);
      if (Actual.width>2) putpixel(x-1,y);
    }

  }

}



void draw_polyline(int N, PointType points[])
{
  if (Actual.hatchDefined || Actual.imageDefined) hatch(N, points);

  int i=0;

  while (i<(N-1))
  {
  draw_line(points[i].x,points[i].y,points[i+1].x,points[i+1].y);
  i++;
  }
  draw_line(points[0].x,points[0].y,points[N-1].x,points[N-1].y);
  
}

void draw_circle(float x,float y,float r)
{
  float alfa,cosa,sina,x1,y1,x2,y2;
  int steps,i;

  alfa=2*acos((r-CIRCLE_Z)/r);
  steps=Fround((2*PI)/alfa);

  cosa=cos((2*PI)/steps);
  sina=sin((2*PI)/steps);
// vytvoreni pole
  i=steps;
  PointType *points = new PointType[steps];

  x1=r;
  y1=0;
  while (i>0)
  {
    x2=x1*cosa-y1*sina;
    y2=x1*sina+y1*cosa;
    points[i-1].x =x+x1;
    points[i-1].y =y+y1;
    //draw_line(x1+x,y1+y,x2+x,y2+y);
    x1=x2;  y1=y2;
    i--;
  }
  draw_polyline(steps,points);
  delete points;
}




void draw_arc(float x,float y,float r,float angle1, float angle2)
{
  float alfa,cosa,sina,dangle,x1,y1,x2,y2,tmp;
  int steps;

  tmp=(angle2-angle1+360)/360.0;
  dangle=(tmp-(int)tmp)*2*PI;


  alfa=2*acos((r-CIRCLE_Z)/r);
  steps=Fround(dangle/alfa);
  alfa=dangle/(float)steps;

  cosa=cos(alfa);
  sina=sin(alfa);
  x1=r*cos((angle1/360.0)*2*PI);
  y1=r*sin((angle1/360.0)*2*PI);
  while (steps>0)
  {
    x2=x1*cosa-y1*sina;
    y2=x1*sina+y1*cosa;
    draw_line(x1+x,y1+y,x2+x,y2+y);
    x1=x2;  y1=y2;
    steps--;
  }
}



void draw_coons(int N, PointType points[])
{
  float x1,y1,x2,y2,t,t2,t3,step,k1,k2,k3,k4;
  int i,p0,p1,p2,p3;
  if (N<4) return;
  step=1/(float)COONS_STEPS;
  N=N-1;
  i=-2;
  while (i<N)
  {


    p0=max(i,0);
    p1=max(i+1,0);
    p2=min(i+2,N);
    p3=min(i+3,N);


    x1=(points[p0].x+4*points[p1].x+points[p2].x)/6;
    y1=(points[p0].y+4*points[p1].y+points[p2].y)/6;

    t=step;
    while (t<(1+step))
    {
      t2=t*t;
      t3=t2*t;
      k1=(-t3+3*t2-3*t+1)/6;
      k2=(3*t3-6*t2+4)/6;
      k3=(-3*t3+3*t2+3*t+1)/6;
      k4=t3/6;

      x2=k1*points[p0].x+k2*points[p1].x+k3*points[p2].x+k4*points[p3].x;
      y2=k1*points[p0].y+k2*points[p1].y+k3*points[p2].y+k4*points[p3].y;


/*      x2=(1/6)*(t3*( -points[p0].x+3*points[p1].x-3*points[p2].x+points[p3].x)+
      	        t2*(3*points[p0].x-6*points[p1].x+3*points[p2].x)+
                t*(-3*points[p0].x+               3*points[p2].x)+
	             (points[p0].x+4*points[p1].x+  points[p2].x));

      y2=(1/6)*(t3*( -points[p0].y+3*points[p1].y-3*points[p2].y+points[p3].y)+
                t2*(3*points[p0].y-6*points[p1].y+3*points[p2].y)+
                t*(-3*points[p0].y+               3*points[p2].y)+
	             (points[p0].y+4*points[p1].y+  points[p2].y));
*/

      draw_line(x1,y1,x2,y2);

      x1=x2; y1=y2;

      t=t+step;

    }

  i++;
  }

}

