#include "svf.h"

/*
 * define the coding of end points
 */
typedef unsigned int code;
enum {TOP = 0x1, BOTTOM = 0x2, RIGHT = 0x4, LEFT = 0x8};


/*
 * compute the code of a point relative to a rectangle
 */

code ComputeCode (double x, double y)
{
        code c = 0;
        if (y > Actual.clip[1].y)
                c |= TOP;
        else if (y < Actual.clip[0].y)
                c |= BOTTOM;
        if (x > Actual.clip[1].x)
                c |= RIGHT;
        else if (x < Actual.clip[0].x)
                c |= LEFT;
        return c;
}

/*
 *      clip line P0(x1,y1)-P1(x1,y2) against a rectangle
 */
void draw_clipped_line(PointCoord x1, PointCoord y1, PointCoord x2, PointCoord y2)
{
        code C0, C1, C;
        double x, y;

        C0 = ComputeCode (x1, y1);
        C1 = ComputeCode (x2, y2);

        for (;;) {
                /* trivial accept: both ends in rectangle */
                if ((C0 | C1) == 0) {
                draw_absolute_line(x1,y1,x2,y2);
                        return;
                }

                /* trivial reject: both ends on the external side of the rectangle */
                if ((C0 & C1) != 0)
                        return;

                /* normal case: clip end outside rectangle */
                C = C0 ? C0 : C1;
                if (C & TOP) {
                        x = x1 + (x2 - x1) * (Actual.clip[1].y - y1) / (y2 - y1);
                        y = Actual.clip[1].y;
                } else if (C & BOTTOM) {
                        x = x1 + (x2 - x1) * (Actual.clip[0].y - y1) / (y2 - y1);
                        y = Actual.clip[0].y;
                } else if (C & RIGHT) {
                        x = Actual.clip[1].x;
                        y = y1 + (y2 - y1) * (Actual.clip[1].x - x1) / (x2 - x1);
                } else {
                        x = Actual.clip[0].x;
                        y = y1 + (y2 - y1) * (Actual.clip[0].x - x1) / (x2 - x1);
                }

                /* set new end point and iterate */
                if (C == C0) {
                        x1 = x; y1 = y;
                        C0 = ComputeCode (x1, y1);
                } else {
                        x2 = x; y2 = y;
                        C1 = ComputeCode (x2, y2);
                }
        }
        
        /* notreached */
}

