
#include "svf.h"

typedef struct {short y, xl, xr, dy;} Segment;
/*
 * Filled horizontal segment of scanline y for xl<=x<=xr.
 * Parent segment was on line y-dy.  dy=1 or -1
 */

#define MAX 100000		/* max depth of stack */

#define PUSH(Y, XL, XR, DY)	/* push new segment on stack */ \
    if (sp<stack+MAX && Y+(DY)>wy0 && Y+(DY)<=wy1) \
    {sp->y = Y; sp->xl = XL; sp->xr = XR; sp->dy = DY; sp++;}

#define POP(Y, XL, XR, DY)	/* pop segment off stack */ \
    {sp--; Y = sp->y+(DY = sp->dy); XL = sp->xl; XR = sp->xr; }

/*
 * fill: set the pixel at (x,y) and all of its 4-connected neighbors
 * with the same pixel value to the new pixel value nv.
 * A 4-connected neighbor is a pixel above, below, left, or right of a pixel.
 */

inline bool compare(const ColorType &a, const ColorType &b) {
    return (a.r==b.r && a.g==b.g && a.b==b.b);
}

inline bool compare(const ColorAlphaType &a, const ColorAlphaType &b) {
    return (a.color.r==b.color.r && a.color.g==b.color.g && a.color.b==b.color.b && a.alpha==b.alpha);
}

#include <stdio.h>

void flood_fill(int x, int y, int wx0, int wy0, int wx1, int wy1)
 {

    int l, x1, x2, dy;
    Segment stack[MAX], *sp = stack;	/* stack of filled segments */

    ColorAlphaType nv;
    nv.color = Actual.color;		   /* read pv at seed point */
    nv.alpha = Actual.alpha;
    ColorAlphaType ov = getpixel(x, y);	   /* read pv at seed point */
    
    if (compare(ov,nv) || x<wx0 || x>wx1 || y<wy0 || y>wy1) return;
    PUSH(y, x, x, 1);			/* needed in some cases */
    PUSH(y+1, x, x, -1);		/* seed segment (popped 1st) */

    int cnt = 0;
    while (sp>stack) {
	cnt++;
//	printf("Flood %d\n",cnt);;
	if (cnt>10000) break;
	/* pop segment off stack and fill a neighboring scan line */
	POP(y, x1, x2, dy);
	/*
	 * segment of scan line y-dy for x1<=x<=x2 was previously filled,
	 * now explore adjacent pixels in scan line y
	 */
	for (x=x1; x>=wx0 && compare(getpixel(x, y), ov); x--)
	    putpixel(x, y);
	if (x>=x1) goto skip;
	l = x+1;
	if (l<x1) PUSH(y, l, x1-1, -dy);		/* leak on left? */
	x = x1+1;
	do {
	    for (; x<=wx1 && compare(getpixel(x, y), ov); x++) putpixel(x, y);
	    PUSH(y, l, x-1, dy);
	    if (x>x2+1) PUSH(y, x2+1, x-1, -dy);	/* leak on right? */
skip:	    

for (x++; x<=x2 && !compare(getpixel(x, y), ov); x++);
	    l = x;
	} while (x<=x2);
    }
}
