
// COMMON TYPES AND DECLARATIONS

struct ColorType {
  int r, g, b;
};

struct ColorAlphaType {
  ColorType color;
  int alpha;
};

typedef float PointCoord;

struct PointType {
  PointCoord x,y;
  PointType operator+(const PointType &point) {
    PointType result;
    result.x = x + point.x;
    result.y = y + point.y;
    return result;
  }
  PointType operator/(int div) {
    PointType result;
    result.x = x/div;
    result.y = y/div;
    return result;
  }
};

struct Matrix {
  struct Column {
    float column[3];
  };
  Column row[3];
  Matrix();
  Matrix& operator*=(const Matrix &matrix);
  void Transform(PointCoord &x, PointCoord &y) const;
  void Clear();
  void Translate(float x, float y);
  void Rotate(float a);
  void Scale(float x, float y);
};

struct Settings {
  ColorType color, hatchcolor;
  int width, imagex, imagey, imagebpp, imageofsx, imageofsy;
  int alpha, alphafunc, alphaupdate;
  unsigned char *imagedata;
  float hatchangle, hatchgap;
  Matrix matrix;
  PointType clip[2];
  bool clipDefined, hatchDefined, imageDefined;
};

// FORWARD DECLARATIONS OF FUNCTIONS

void setcolor(int r, int g, int b);
void setalpha(int alpha);

void putpixel(int x, int y);
ColorAlphaType getpixel(int x, int y);

void draw_absolute_line(PointCoord x1, PointCoord y1, PointCoord x2, PointCoord y2);
void draw_line(PointCoord x1, PointCoord y1, PointCoord x2, PointCoord y2);
void draw_polyline(int n, PointType points[]);
void draw_circle(PointCoord x,PointCoord y,PointCoord r);

void draw_arc(PointCoord x,PointCoord y,PointCoord r,float angle1, float angle2);
void draw_bezier(int n, PointType points[]);
void draw_coons(int n, PointType points[]);

void draw_clipped_line(PointCoord x1, PointCoord y1, PointCoord x2, PointCoord y2);

void hatch(int n, PointType points[]);
unsigned char *load_tga(const char *name, int &sizex, int &sizey, int &bpp);

void flood_fill(int x, int y, int x0, int y0, int x1, int y1);

ColorAlphaType mix(ColorAlphaType fg, ColorAlphaType bg);

// GLOBAL SYMBOLS

extern Settings Actual;
extern float Factor;
extern PointType Offset;

// COMMON TYPES AND DECLARATIONS

struct ColorType {
  int r, g, b;
};

struct ColorAlphaType {
  ColorType color;
  int alpha;
};

typedef float PointCoord;

struct PointType {
  PointCoord x,y;
  PointType operator+(const PointType &point) {
    PointType result;
    result.x = x + point.x;
    result.y = y + point.y;
    return result;
  }
  PointType operator/(int div) {
    PointType result;
    result.x = x/div;
    result.y = y/div;
    return result;
  }
};

struct Matrix {
  struct Column {
    float column[3];
  };
  Column row[3];
  Matrix();
  Matrix& operator*=(const Matrix &matrix);
  void Transform(PointCoord &x, PointCoord &y) const;
  void Clear();
  void Translate(float x, float y);
  void Rotate(float a);
  void Scale(float x, float y);
};

struct Settings {
  ColorType color, hatchcolor;
  int width, imagex, imagey, imagebpp, imageofsx, imageofsy;
  int alpha, alphafunc, alphaupdate;
  unsigned char *imagedata;
  float hatchangle, hatchgap;
  Matrix matrix;
  PointType clip[2];
  bool clipDefined, hatchDefined, imageDefined;
};

// FORWARD DECLARATIONS OF FUNCTIONS

void setcolor(int r, int g, int b);
void setalpha(int alpha);

void putpixel(int x, int y);
ColorAlphaType getpixel(int x, int y);

void draw_absolute_line(PointCoord x1, PointCoord y1, PointCoord x2, PointCoord y2);
void draw_line(PointCoord x1, PointCoord y1, PointCoord x2, PointCoord y2);
void draw_polyline(int n, PointType points[]);
void draw_circle(PointCoord x,PointCoord y,PointCoord r);

void draw_arc(PointCoord x,PointCoord y,PointCoord r,float angle1, float angle2);
void draw_bezier(int n, PointType points[]);
void draw_coons(int n, PointType points[]);

void draw_clipped_line(PointCoord x1, PointCoord y1, PointCoord x2, PointCoord y2);

void hatch(int n, PointType points[]);
unsigned char *load_tga(const char *name, int &sizex, int &sizey, int &bpp);

void flood_fill(int x, int y, int x0, int y0, int x1, int y1);

ColorAlphaType mix(ColorAlphaType fg, ColorAlphaType bg);

// GLOBAL SYMBOLS

extern Settings Actual;
extern float Factor;
extern PointType Offset;
