
#include "svf.h"

PointCoord abs(PointCoord x) {
  return (x>0) ? x : -x;
}

const int PRECISION = 10;

static void casteljau_divide(PointType orig[], PointType left[], PointType right[]) {
  left[0]  = orig[0];
  left[1]  = (orig[0]+orig[1])/2;
  PointType tmp = (orig[1]+orig[2])/2;
  left[2]  = (left[1]+tmp)/2;
  right[3] = orig[3];
  right[2] = (orig[2]+orig[3])/2;
  right[1] = (tmp+right[2])/2;
  left[3]  = right[0] = (left[2]+right[1])/2;
}

static int area(const PointType &a, const PointType &b, const PointType &c) {
  return abs((b.x-a.x)*(c.y-a.y)-(b.y-a.y)*(c.x-a.x));
}

static bool flat_enough(const PointType points[]) {
  int result = area(points[0],points[1],points[3]) + area(points[1],points[2],points[3]);
  return result <= PRECISION;
}
 
static void draw_bezier(PointType points[])
 {
  if (flat_enough(points)) {
    draw_line(points[0].x, points[0].y, points[3].x, points[3].y);
  }
  else {
    PointType left[4], right[4];
    casteljau_divide(points,left,right);
    draw_bezier(4,left);
    draw_bezier(4,right);
  }
}

void draw_bezier(int n, PointType points[])
 {
  int i;
  for (i=0; i+3<n; i+=3) draw_bezier(points+i);
  PointType last[4];
  int base = i;
  if (base+1 < n) {
    last[0] = (base+0)<n ? points[base+0] : points[n-1];
    last[1] = (base+1)<n ? points[base+1] : points[n-1];
    last[2] = (base+2)<n ? points[base+2] : points[n-1];
    last[3] = (base+3)<n ? points[base+3] : points[n-1];
    draw_bezier(last);
  }
}

