
#include <qpainter.h>
#include <qimage.h>
#include <qapplication.h>
#include <qwidget.h>
#include <stdio.h>

#include <stdlib.h>

#include "svf.h"

const int BORDER = 10;

Settings Actual,Saved;
float Factor;
PointType TopRight,BottomLeft,Size,Offset;
enum { Draw, BoundBox } Mode;
struct {
  char name[256];
  unsigned char *data;
  int x,y;
} image[16];
int images = 0;
int InitialAlpha = 255;

ColorAlphaType *buffer = NULL;
int bufferw, bufferh;

void setcolor(int r, int g, int b) {
  Actual.color.r = r;
  Actual.color.g = g;
  Actual.color.b = b;
}

void setalpha(int alpha) {
  Actual.alpha = alpha;
}

void setwidth(int width) {
  Actual.width = width;
}

void push() {
  Saved = Actual;
}

void pop() {
  Actual = Saved;
}

void pr(ColorAlphaType x) {
  printf("[%d %d %d %d] ",x.color.r,x.color.g,x.color.b,x.alpha);
}

bool invalid(int x, int y) {
    return (((int)Offset.x+x<0) || ((int)Offset.x+x>=bufferw)
         || ((int)Offset.y-y<0) || ((int)Offset.y-y>=bufferh));
}

void putpixel(int x, int y) {
  if (Mode==Draw) {
    if (invalid(x,y)) { printf("%d,%d not valid in putpixel\n",x,y);
     exit(3); }
    ColorAlphaType &act = buffer[bufferw*((int)Offset.y-y)+(int)Offset.x+x];
    ColorAlphaType bg = act;
    ColorAlphaType fg;
    fg.color = Actual.color;                                                    
    fg.alpha = Actual.alpha;    
    act = mix(fg,bg);
  }
  else {
    if (x>TopRight.x) TopRight.x = x;
    if (y>TopRight.y) TopRight.y = y;
    if (x<BottomLeft.x) BottomLeft.x = x;
    if (y<BottomLeft.y) BottomLeft.y = y;
  }
}

ColorAlphaType getpixel(int x, int y) {
  if (invalid(x,y)) { /* printf("%d,%d not valid in getpixel\n",x,y); */
  /*exit(3);*/
    ColorAlphaType out;
    out.color.r = 152;
    out.color.g = 13;
    out.color.b = 6;
    out.alpha = 253;
    return out;
   }
  ColorAlphaType &act = buffer[bufferw*((int)Offset.y-y)+(int)Offset.x+x];
  return act;
}

void LoadFile(char* &buffer, long &size, const char *name) {
  FILE* file = fopen(name,"rb");
  if (!file) { printf("%s not found\n",name); exit(1); }
  fseek(file,0,SEEK_END); size = ftell(file); fseek(file,0,SEEK_SET);
  buffer = new char[size];
  fread(buffer,1,size,file);
  fclose(file);
}

PointType *ParsePoints(int n, const char *ptr) {
  PointType *points = new PointType[n];
  for (int i=0; i<n; i++) {
    float x,y; int pos; 
    sscanf(ptr,"%f %f%n",&x,&y,&pos);
    points[i].x = Factor*x; points[i].y = Factor*y;
    ptr += pos;
  }
  return points;
}

void ParseFile(const char *buffer, int size) {
  Actual.matrix.Clear();
  Actual.matrix.row[0].column[0] = 1;
  Actual.matrix.row[1].column[1] = 1;
  Actual.matrix.row[2].column[2] = 1;
  Actual.clipDefined = false;
  Actual.hatchDefined = false;
  Actual.alpha = 255;
  Actual.alphafunc = 0;
  Actual.alphaupdate = 1;
  const char *line = buffer;
  const char *stop = buffer+size;
  while (line<stop) {
    float x,y,x1,y1,x2,y2,rad,a,a1,a2,gap;
    int pos,n,r,g,b,alpha,func,update;
    char name[256];
    if (line[0] == '#');
    else if (line[0] == '{') push();
    else if (line[0] == '}') pop();
    else if (sscanf(line,"L %f %f %f %f",&x1,&y1,&x2,&y2) == 4) {
      draw_line(Factor*x1,Factor*y1,Factor*x2,Factor*y2);
    }
    else if (sscanf(line,"P %d%n",&n,&pos) >= 1) {
      PointType *points = ParsePoints(n,line+pos);
      draw_polyline(n,points);
      delete points;
    }
    else if (sscanf(line,"B %d%n",&n,&pos) >= 1) {
      PointType *points = ParsePoints(n,line+pos);
      draw_bezier(n,points);
      delete points;
    }
    else if (sscanf(line,"K %d%n",&n,&pos) >= 1) {
      PointType *points = ParsePoints(n,line+pos);
      draw_coons(n,points);
      delete points;
    }
    else if (sscanf(line,"C %f %f %f",&x,&y,&rad) == 3) {
      draw_circle(Factor*x,Factor*y,Factor*rad);
    }
    else if (sscanf(line,"A %f %f %f %f %f",&x,&y,&rad,&a1,&a2) == 5) {
      draw_arc(Factor*x,Factor*y,Factor*rad,a1,a2);
    }
    else if (sscanf(line,"W %d",&n) == 1) {
      setwidth(n);
    }
    else if (sscanf(line,"T %f %f",&x,&y) == 2) {
      Matrix translate;
      translate.Translate(Factor*x,Factor*y);
      Actual.matrix *= translate;
    }
    else if (sscanf(line,"R %f %f %f",&x,&y,&a) == 3) {
      Matrix translate1; translate1.Translate(-Factor*x,-Factor*y);
      Matrix rotate; rotate.Rotate(a);
      Matrix translate2; translate2.Translate(Factor*x,Factor*y);
      Actual.matrix *= translate1;
      Actual.matrix *= rotate;
      Actual.matrix *= translate2;
    }
    else if (sscanf(line,"S %f %f",&x,&y) == 2) {
      Matrix scale;
      scale.Scale(x,y);
      Actual.matrix *= scale;
    }
    else if (sscanf(line,"O %d %d %d",&r,&g,&b) == 3) {
      setcolor(r,g,b);
    }
    else if (sscanf(line,"N %f %f %f %f",&x1,&y1,&x2,&y2) == 4) {
      Actual.clip[0].x = Factor*x1;
      Actual.clip[0].y = Factor*y1;
      Actual.clip[1].x = Factor*x2;
      Actual.clip[1].y = Factor*y2;
      Actual.clipDefined = true;
    }
    else if (sscanf(line,"H %f %f %d %d %d",&a,&gap,&r,&g,&b) == 5) {
      Actual.hatchangle = a;
      Actual.hatchgap = gap;
      Actual.hatchcolor.r = r;
      Actual.hatchcolor.g = g;
      Actual.hatchcolor.b = b;
      Actual.hatchDefined = true;
      Actual.imageDefined = false;
    }
    else if (sscanf(line,"I %f %f %s",&x,&y,name) == 3) {
      if (Mode==BoundBox) {
        unsigned char *data = load_tga(name,Actual.imagex,Actual.imagey,Actual.imagebpp);
	if (!data)
	  printf("Can't load image\n");
        else {
	  image[images].data = data;
    	  image[images].x = Actual.imagex;
          image[images].y = Actual.imagey;
  	  strcpy(image[images].name,name);
	  images++;
	}
	Actual.imagedata = NULL;
      }
      else {
        for (int i=0; i<images; i++)
	  if (0==strcmp(name,image[i].name)) {
	    Actual.imagedata = image[i].data;
	    Actual.imagex    = image[i].x;
	    Actual.imagey    = image[i].y;
	    break;
	  }
        }
      Actual.hatchDefined = false;
      Actual.imageDefined = true;
      Actual.hatchangle = 0;
      Actual.hatchgap = 1;
      Actual.imageofsx = x;
      Actual.imageofsy = y;
    }
    else if (sscanf(line,"F %f %f %d %d %d",&x,&y,&r,&g,&b) == 5) {
      PointCoord cx = Factor*x;
      PointCoord cy = Factor*y;
      ColorType s = Actual.color;
      setcolor(r,g,b);
//      Actual.matrix.Transform(cx,cy);
      if (Mode==Draw)
        flood_fill(cx,cy,-Offset.x,Offset.y-bufferh,bufferw-Offset.x,Offset.y);
      Actual.color = s;
    }
    else if (sscanf(line,"D %d %d %d",&alpha,&func,&update) == 3) {
      Actual.alpha       = alpha;
      Actual.alphafunc   = func;
      Actual.alphaupdate = update;
    }
    else if (line[0] == 'E') {
      break;
    }
    else {
      printf("Unrecognized line (%c)\n",line[0]);
    }
    while ((*line!='\x0D' && *line!='\x0A') && line<stop) line++;
    while ((*line=='\x0D' || *line=='\x0A') && line<stop) line++;
  }
}

char *FileBuffer; long FileSize;

class SVFDisplay : public QWidget {
public:
  SVFDisplay() {
#if QT_VERSION >= 300
    setEraseColor(Qt::black);
#endif
  }
  virtual QSize sizeHint() const {
    return QSize(640,480);
  }
protected:
  void paintEvent(QPaintEvent *) {
    int w = bufferw = width();
    int h = bufferh = height(); 
    if (buffer) delete buffer;
    buffer = new ColorAlphaType[bufferw*bufferh];
    ColorAlphaType initial;
    initial.color.r = initial.color.g = initial.color.b = 0;
    initial.alpha = InitialAlpha;
    for (int y=0; y<bufferh; y++) for (int x=0; x<bufferw; x++) {
      buffer[bufferw*y+x] = initial;
    }
    PointType window = { w-2*BORDER, h-2*BORDER };
    float FactorX = (float)window.x/Size.x;
    float FactorY = (float)window.y/Size.y;
    Mode = Draw; Factor = (FactorX>FactorY) ? FactorY : FactorX;
    Offset.x = (window.x-Factor*Size.x)/2+BORDER-Factor*BottomLeft.x;
    Offset.y = (window.y-Factor*Size.y)/2+BORDER+Factor*BottomLeft.y+Factor*Size.y;
    ParseFile(FileBuffer,FileSize);
    QPainter paint(this);
    for (int y=0; y<bufferh; y++) for (int x=0; x<bufferw; x++) {
      ColorAlphaType &act = buffer[bufferw*y+x];
      paint.setPen(QColor(act.color.r,act.color.g,act.color.b));
      paint.drawPoint(x,y);
    }
  }
  void mousePressEvent(QMouseEvent *e) {
    if (e->button()==LeftButton) {
      setcolor(255,255,0);
      Actual.alpha = 255;
      Actual.alphafunc = 0;
      Actual.alphaupdate = 1;
      flood_fill(e->pos().x()-Offset.x,Offset.y-e->pos().y(),
  	         -Offset.x,Offset.y-bufferh,bufferw-Offset.x,Offset.y);
      QPainter paint(this);
      for (int y=0; y<bufferh; y++) for (int x=0; x<bufferw; x++) {
        ColorAlphaType &act = buffer[bufferw*y+x];
        paint.setPen(QColor(act.color.r,act.color.g,act.color.b));
        paint.drawPoint(x,y);
      }
    }
    else {
      int x = e->pos().x();
      int y = e->pos().y();
      ColorAlphaType &act = buffer[bufferw*y+x];
      char tmp[256];
      sprintf(tmp,"color: %d %d %d alpha: %d",act.color.r,act.color.g,act.color.b,act.alpha);
      setCaption(tmp);
    }
  }
};

int view(const char* file) {
  // LOAD FILE
  LoadFile(FileBuffer,FileSize,file);

  // DETERMINE BOUNDING BOX
  TopRight.x = TopRight.y = -99999999;
  BottomLeft.x = BottomLeft.y = +99999999;
  Mode = BoundBox; Factor = 1;
  ParseFile(FileBuffer,FileSize);
  Size.x = TopRight.x - BottomLeft.x;
  Size.y = TopRight.y - BottomLeft.y;

  // CREATE WINDOW AND RUN APPLICATION
  SVFDisplay svf;
  qApp->setMainWidget(&svf);
  char title[1024];
  sprintf(title,"%s - SVFViewer",file);
  svf.setCaption(title);
  svf.show();
  int result = qApp->exec();
  delete FileBuffer;
  return result;
}

int main(int argc, char **argv) {
  QApplication a(argc, argv);
  InitialAlpha = 255;
  for (int param=1; param<argc; param++) {
    int integer;
    if (param==1 && sscanf(argv[param],"%d",&integer) == 1) {
      InitialAlpha = integer;
    }
    else {
      int result = view(argv[param]);
      if (result) return result;
    }
  }
  return 0;
}
