
#include "svf.h"
#include <stdlib.h>

#include <stdio.h>

#define max(x1,x2) ((x1 > x2) ? x1 : x2)
#define min(x1,x2) ((x1 > x2) ? x2 : x1)

struct Edge {
  int ymin, ymax;
  float x, m;
  PointType first;
};

int compare_ymin(const void *a, const void *b) {
  if (((Edge*)a)->ymin < ((Edge*)b)->ymin) return -1;
  if (((Edge*)a)->ymin > ((Edge*)b)->ymin) return 1;
  return 0;
}

int compare_x(const void *a, const void *b) {
  if (((Edge*)a)->x < ((Edge*)b)->x) return -1;
  if (((Edge*)a)->x > ((Edge*)b)->x) return 1;
  return 0;
}

void hatch(int n, PointType points[]) {
  Matrix rotate;
  rotate.Rotate(-Actual.hatchangle);
  int ymin = +999999999;
  int ymax = -999999999;
  Edge *edge = new Edge[n];
  int edges = 0;
  for (int i=0; i<n; i++) {
    PointType first  = points[i];
    PointType second = points[(i+1) % n];
    Actual.matrix.Transform(first.x,first.y);
    Actual.matrix.Transform(second.x,second.y);
    rotate.Transform(first.x,first.y);
    rotate.Transform(second.x,second.y);
    if (first.y<ymin) ymin = first.y;
    if (first.y>ymax) ymax = first.y;
    if (first.y==second.y) continue;
    edge[edges].first = first;
    edge[edges].m = (second.x-first.x)/(second.y-first.y);
    edge[edges].ymin = min(first.y,second.y)+1;
    edge[edges].ymax = max(first.y,second.y);
    edges++;
  }
  qsort(edge,edges,sizeof(Edge),&compare_ymin);

  Edge *active = new Edge[edges];
  int gap = Factor*Actual.hatchgap;
  if (Actual.imageDefined) {
    gap = 1;
  }
  for (int y=ymin; y<ymax; y+=gap) {
    int actives = 0;
    for (int i=0; i<edges; i++) {
      Edge &act = edge[i];
      if (act.ymin<=y && y<=act.ymax) {
        act.x = act.m*(y-act.first.y) + act.first.x;
        active[actives++] = act;
      }
    }
    qsort(active,actives,sizeof(Edge),&compare_x);
    Matrix rotate;
    rotate.Rotate(+Actual.hatchangle);
    Settings saved = Actual; Actual.color = Actual.hatchcolor; Actual.width = 1;
    for (int i=0; i+1<actives; i+=2) {
      PointCoord x1,y1,x2,y2;
      x1 = active[i].x;
      x2 = active[i+1].x;
      y1 = y2 = y;
      if (Actual.hatchDefined) {
        rotate.Transform(x1,y1);
        rotate.Transform(x2,y2);
        draw_absolute_line(x1,y1,x2,y2);
      }
      else if (Actual.imagedata) {
        for (int x=x1; x<=x2; x++) {
          int ofsx = -Actual.imageofsx*Factor;
	  int ofsy = -Actual.imageofsy*Factor;
	  int offset =
	    (ofsx+x+1000*Actual.imagex)%Actual.imagex
	    +
	    Actual.imagex*(Actual.imagey-1-(ofsy+y+1000*Actual.imagey)%Actual.imagey);
	  offset *=  Actual.imagebpp/8;
	  unsigned char *ptr = Actual.imagedata + offset;
	  if (Actual.clipDefined) {
            if (x < Actual.clip[0].x) continue;
            if (x > Actual.clip[1].x) continue;
            if (y < Actual.clip[0].y) continue;
            if (y > Actual.clip[1].y) continue;
	  }
	  setcolor(ptr[2],ptr[1],ptr[0]);
	  if (Actual.imagebpp==32) setalpha(ptr[3]);
	  putpixel(x,y);
	}
      }
    }
    Actual = saved;
  }
  delete active;
  delete edge;
}

unsigned char *load_tga(const char *name, int &sizex, int &sizey, int &bpp) {
  FILE *file = fopen(name,"rb");
  if (!file) return NULL;
  struct { char head[8]; short offset[2];
           unsigned short x,y; unsigned char bpp, place;} header;
  fread(&header,1,sizeof(header),file);
  int orientation = header.place & 0x30;
  sizex = header.x;
  sizey = header.y;
  bpp   = header.bpp;
  long size = sizex*sizey*bpp/8;
  unsigned char *data = new unsigned char[size];
  fread(data,1,size,file);
  if (orientation==0) {
    for (int y=0; y<sizey/2; y++) {
      for (int x=0; x<sizex*header.bpp/8; x++) {
        int offset1 = (y*sizex*header.bpp/8)+x;
        int offset2 = ((sizey-y-1)*sizex*header.bpp/8)+x;
        unsigned char save = data[offset1];
        data[offset1] = data[offset2];
	data[offset2] = save;
      }
    }
  }
  fclose(file);
  return data;
}
