
#include "svf.h"
#include "string.h"
#include <math.h>

const float DEG2RAD = 0.017453292519943;

void draw_line(PointCoord x1, PointCoord y1, PointCoord x2, PointCoord y2) {
  Actual.matrix.Transform(x1,y1);
  Actual.matrix.Transform(x2,y2);

  if (Actual.clipDefined) draw_clipped_line(x1,y1,x2,y2);
  else draw_absolute_line(x1,y1,x2,y2);
}

Matrix::Matrix() {
  Clear();
}

Matrix& Matrix::operator*=(const Matrix &matrix) {
  Matrix result;
  for (int r=0; r<3; r++)
    for (int c=0; c<3; c++) {
      float &act = result.row[r].column[c];
      act = 0;
      for (int i=0; i<3; i++)
        act += row[r].column[i]*matrix.row[i].column[c];
    }
  *this = result;
  return *this;
}

void Matrix::Transform(PointCoord &x, PointCoord &y) const {
  PointCoord _x = x*row[0].column[0] + y*row[1].column[0] + row[2].column[0];
  PointCoord _y = x*row[0].column[1] + y*row[1].column[1] + row[2].column[1];
  x = _x; y = _y;
}

void Matrix::Clear() {
  memset(row,0,3*sizeof(Column));
}

void Matrix::Translate(float x, float y) {
  Clear();
  row[0].column[0] = 1;
  row[1].column[1] = 1;
  row[2].column[0] = x;
  row[2].column[1] = y;
  row[2].column[2] = 1;
}

void Matrix::Rotate(float a) {
  Clear();
  a *= DEG2RAD;
  row[0].column[0] = cos(a);
  row[0].column[1] = sin(a);
  row[1].column[0] = -sin(a);
  row[1].column[1] = cos(a);
  row[2].column[2] = 1;
}

void Matrix::Scale(float x, float y) {
  Clear();
  row[0].column[0] = x;
  row[1].column[1] = y;
  row[2].column[2] = 1;
}
